from tuoni.TuoniExceptions import *

class TuoniListener:
    def __init__(self, conf, c2):
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.listener_id = conf["id"]
        self.name = conf["name"]
        self.info = conf["info"]
        self.status = conf["status"]
        self.plugin = conf["plugin"]
        self.configuration = conf["configuration"]

    def stop(self):
        if self.listener_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_put("/api/v1/listeners/%s/stop" % self.listener_id)
        self._load_conf(data)

    def start(self):
        if self.listener_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_put("/api/v1/listeners/%d/stop" % self.listener_id)
        self._load_conf(data)

    def delete(self):
        if self.listener_id is None:
            raise ExceptionTuoniDeleted("")
        self.c2.request_delete("/api/v1/listeners/%d" % self.listener_id)
        self.listener_id = None





import time
import requests
import json
import threading
import shutil
from tuoni.TuoniExceptions import *
from tuoni.TuoniListenerPlugin import *
from tuoni.TuoniListener import *
from tuoni.TuoniAgent import *
from tuoni.TuoniPayloadPlugin import *


class TuoniC2:
    def __init__(self):
        self.token = None
        self.url = None
        self.monitoring_threads = []

    def login(self, url, username, password):
        headers = {
            "Authorization":
                "Basic " +
                base64
                .b64encode(
                    (username + ":" + password)
                    .encode('utf-8')
                )
                .decode('utf-8')
        }

        response = requests.request("POST", "%s/api/v1/auth/login" % url, headers=headers, verify=False)
        if response.status_code != 200:
            raise ExceptionTuoniAuthentication(response.text)
        self.token = response.text
        self.url = url

    def _request_check(self):
        if self.token is None:
            raise ExceptionTuoniAuthentication("You have not done the login")

    def _raise_request_exception(result_msg):
        msg = result_msg
        try:
            data = json.loads(result_msg)
            msg = data["message"]
        except:
            pass
        raise ExceptionTuoniRequestFailed(msg)

    def request_get(self, uri, result_as_json=True):
        self._request_check()
        headers = {"Authorization": "Bearer " + self.token}
        response = requests.get("%s%s" % (self.url, uri), headers=headers, verify=False)
        if response.status_code != 200:
            TuoniC2._raise_request_exception(response.text)
        if response.text == "":
            return None
        if result_as_json:
            return json.loads(response.text)
        return response.text

    def request_get_file(self, uri, file_name):
        self._request_check()
        headers = {"Authorization": "Bearer " + self.token}
        response = requests.get("%s%s" % (self.url, uri), headers=headers, verify=False, stream=True)
        if response.status_code != 200:
            TuoniC2._raise_request_exception(response.text)
        with open(file_name, 'wb') as out_file:
            shutil.copyfileobj(response.raw, out_file)
        del response

    def request_post(self, uri, json_data = None, files = None):
        self._request_check()
        headers = {"Authorization": "Bearer " + self.token}
        response = None
        if files is None:
            response = requests.post("%s%s" % (self.url, uri), headers=headers, verify=False, json=json_data)
        else:
            all_data = {}
            if json_data is not None:
                all_data["requestBody"] = (None, json.dumps(json_data), 'application/json')
            for var_name in files:
                all_data[var_name] = (files[var_name][0], files[var_name][1], 'application/octet-stream')
            response = requests.post("%s%s" % (self.url, uri), headers=headers, verify=False, files=all_data)
        if response.status_code != 200:
            TuoniC2._raise_request_exception(response.text)
        if response.text == "":
            return None
        return json.loads(response.text)

    def request_put(self, uri, json_data = None):
        self._request_check()
        headers = {"Authorization": "Bearer " + self.token}
        response = requests.put("%s%s" % (self.url, uri), headers=headers, verify=False, json=json_data)
        if response.status_code != 200:
            TuoniC2._raise_request_exception(response.text)
        if response.text == "":
            return None
        return json.loads(response.text)

    def request_delete(self, uri, json_data = None):
        self._request_check()
        headers = {"Authorization": "Bearer " + self.token}
        response = requests.delete("%s%s" % (self.url, uri), headers=headers, verify=False, json=json_data)
        if response.status_code != 200:
            TuoniC2._raise_request_exception(response.text)
        if response.text == "":
            return None
        return json.loads(response.text)

    def load_listener_plugins(self):
        plugins_data = self.request_get("/api/v1/plugins/listeners")
        plugins = {}
        for plugin_name in plugins_data:
            plugin_data = plugins_data[plugin_name]
            plugin_obj = TuoniListenerPlugin(plugin_data, self)
            plugins[plugin_obj.plugin_id] = plugin_obj
        return plugins

    def load_listeners(self):
        listeners_data = self.request_get("/api/v1/listeners")
        listeners = []
        for listener_id in listeners_data:
            listener_data = listeners_data[listener_id]
            listener_obj = TuoniListener(listener_data, self)
            listeners.append(listener_obj)
        return listeners

    def load_payload_plugins(self):
        plugins_data = self.request_get("/api/v1/plugins/payloads")
        plugins = {}
        for plugin_name in plugins_data:
            plugin_data = plugins_data[plugin_name]
            plugin_obj = TuoniPayloadPlugin(plugin_data, self)
            plugins[plugin_obj.plugin_id] = plugin_obj
        return plugins

    def create_payload(self, payload_template, payload_listener, payload_conf, encrypted = True):
        json_data = {
            "payloadTemplateId": payload_template,
            "configuration": payload_conf,
            "listenerId": payload_listener,
            "encrypted": encrypted
        }
        payload_data = self.request_post("/api/v1/payloads", json_data)
        return payload_data["id"]
        
    def download_payload(self, payload_id, file_name):
        self.request_get_file("/api/v1/payloads/%d/download" % (payload_id), file_name)

    def load_agents(self):
        agents_data = self.request_get("/api/v1/agents")
        agents = []
        for agent_data in agents_data:
            agent_obj = TuoniAgent(agent_data, self)
            agents.append(agent_obj)
        return agents

    def wait_new_agent(self, interval = 1, max_wait = 0):
        agents_data = self.request_get("/api/v1/agents")
        original_agents = []
        for agent_data in agents_data:
            original_agents.append(agent_data["guid"])
        while True:
            time.sleep(interval)
            agents_data = self.request_get("/api/v1/agents")
            for agent_data in agents_data:
                if agent_data["guid"] not in original_agents:
                    return TuoniAgent(agent_data, self)
            if max_wait > 0:
                max_wait -= interval
                if max_wait <= 0:
                    return None

    def on_new_agent(self, function, interval = 1):
        monitor_thread = threading.Thread(target=self._monitor_for_new_agents, args=(function, interval, ), daemon=True)
        monitor_thread.start()
        self.monitoring_threads.append(monitor_thread)

    def _monitor_for_new_agents(self, function, interval=1):
        agents_data = self.request_get("/api/v1/agents")
        original_agents = []
        for agent_data in agents_data:
            original_agents.append(agent_data["guid"])
        while True:
            time.sleep(interval)
            agents_data = self.request_get("/api/v1/agents")
            for agent_data in agents_data:
                if agent_data["guid"] not in original_agents:
                    original_agents.append(agent_data["guid"])
                    agent = TuoniAgent(agent_data, self)
                    agent_event_thread = threading.Thread(target=function, args=(agent, ))
                    agent_event_thread.start()

    def let_it_run(self):
        for monitoring_thread in self.monitoring_threads:
            monitoring_thread.join()



from tuoni.TuoniExceptions import *
from tuoni.TuoniCommand import *
from tuoni.TuoniDefaultCommands import *

class TuoniAgent:
    def __init__(self, conf, c2):
        self.c2 = c2
        self._load_conf(conf)

    def _load_conf(self, conf):
        self.guid = conf["guid"]
        self.first_registration_time = conf["firstRegistrationTime"]
        self.last_callback_time = conf["lastCallbackTime"]
        self.metadata = conf["metadata"]
        self.active = conf["active"]
        self.recentListeners = conf["recentListeners"]
        self.fill_available_commands(conf["availableCommandTemplates"])

    def send_command(self, command_type, command_conf=None, execution_conf = None, files = None):
        if self.guid is None:
            raise ExceptionTuoniDeleted("")
        if isinstance(command_type, TuoniDefaultCommand):
            command_conf = command_type.command_conf
            execution_conf = command_type.execution_conf
            files = command_type.files
            command_type = command_type.command_type
        if command_conf is None:
            command_conf = {}
        data = {
            "template": command_type,
            "configuration": command_conf
        }
        if execution_conf is not None:
            data["execConf"] = execution_conf
        data = self.c2.request_post("/api/v1/agents/%s/commands" % self.guid, data, files)
        return TuoniCommand(data, self.c2)

    def get_commands(self):
        if self.guid is None:
            raise ExceptionTuoniDeleted("")
        commands_data = self.c2.request_get("/api/v1/agents/%s/commands" % self.guid)
        commands = []
        for command_nr in commands_data:
            command_data = commands_data[command_nr]
            command_obj = TuoniCommand(command_data, self.c2)
            commands.append(command_obj)
        return commands

    def delete(self):
        if self.guid is None:
            raise ExceptionTuoniDeleted("")
        self.c2.request_delete("/api/v1/agents/%s" % self.guid)
        self.listener_id = None
        
    def fill_available_commands(self, command_list):
        self.availableCommands = {}
        for cmd in self.c2.request_get("/api/v1/command-templates"):
            if cmd["id"] in command_list:
                self.availableCommands[cmd["name"]] = cmd["id"]
            




import random
from tuoni.TuoniListener import *

class TuoniListenerPlugin:
    def __init__(self, conf, c2):
        self.name = conf["info"]["name"]
        self.vendor = conf["info"]["vendor"]
        self.description = conf["info"]["description"]
        self.plugin_id = conf["identifier"]["id"]
        self.conf_schema = conf["configurationSchema"]
        self.conf_examples = {}
        if "defaultConfiguration" in conf:
            self.conf_examples["default"] = conf["defaultConfiguration"]
        if "exampleConfigurations" in conf:
            for example in conf["exampleConfigurations"]:
                self.conf_examples[example["name"]] = example["configuration"]
        self.c2 = c2

    def create(self, new_listener_conf, new_listener_name=None):
        json_data = {
            "plugin": self.plugin_id,
            "configuration": new_listener_conf
        }
        if new_listener_name is not None:
            json_data["name"] = new_listener_name
        listener_data = self.c2.request_post("/api/v1/listeners", json_data)
        listener_obj = TuoniListener(listener_data, self.c2)
        return listener_obj

    def get_default_conf(self):
        return self.conf_default

    def get_minimal_conf(self):
        return self.conf_default




import json
from tuoni.TuoniExceptions import *


class TuoniResultPart:
    def __init__(self, conf, c2):
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.type = conf["type"]
        if self.type == "text":
            self.name = conf["name"]
            self.value = conf["value"]
        if self.type == "file":
            self.filename = conf["filename"]
            self.href = conf["href"]
            if "//" in self.href:
                self.uri = self.href[self.href.find("//")+2:]
                self.uri = self.uri[self.uri.find("/"):]
            else:
                self.uri = self.href

    def get_as_text(self):
        if self.type == "text":
            return self.value
        if self.type == "file":
            return self.c2.request_get(self.uri, result_as_json=False)
        return None

    def get_as_file(self, filename):
        if self.type == "file":
            self.c2.request_get_file(self.uri, filename)
        return None

    def get_as_json(self):
        if self.type == "text":
            return json.loads(self.value)
        if self.type == "file":
            return json.loads(self.c2.request_get(self.uri, result_as_json=False))
        return None






class ExceptionTuoniAuthentication(Exception):
    def __init__(self, message=None):
        self.message = message
        super().__init__(message)

class ExceptionTuoniRequestFailed(Exception):
    def __init__(self, message=None):
        self.message = message
        super().__init__(message)

class ExceptionTuoniDeleted(Exception):
    def __init__(self, message=None):
        self.message = message
        super().__init__(message)


from tuoni.TuoniResultPart import *


class TuoniResult:
    def __init__(self, conf, c2):
        self.c2 = c2
        self._load_conf(conf)

    def _load_conf(self, conf):
        self.status = conf["status"]
        self.error_message = conf["errorMessage"]
        self.received = conf["received"]
        self.parts = []
        for part_conf in conf["childResults"]:
            part_obj = TuoniResultPart(part_conf, self.c2)
            self.parts.append(part_obj)





import base64

class TuoniDefaultCommand:
    def __init__(self, command_type, command_conf):
        self.command_type = command_type
        self.command_conf = command_conf
        self.execution_conf = None
        self.files = None

class TuoniDefaultPluginCommand(TuoniDefaultCommand):
    def __init__(self, command_type, command_conf, execution_conf = None):
        super().__init__(command_type, command_conf)
        if isinstance(execution_conf, ExecutionNew):
            self.execution_conf = {
                "execType": "NEW",
                "executable": execution_conf.proc_name,
                "suspended": execution_conf.suspended
            }
        elif isinstance(execution_conf, ExecutionExisting):
            self.execution_conf = {
                "execType": "EXISTING",
                "pid": execution_conf.pid
            }
        else:
            self.execution_conf = execution_conf


class ExecutionNew:
    def __init__(self, proc_name="notepad.exe", suspended=True):
        self.proc_name = proc_name
        self.suspended = suspended

class ExecutionExisting:
    def __init__(self, pid):
        self.pid = pid


#########################
## Native commands
#########################
class TuoniCommandBof(TuoniDefaultCommand):
    def __init__(self, bof_file, method =  "go", inputArgs = None, inputArgsEncoding = None, inputAsBytes = None):
        super().__init__("bof", {"method": method, "inputArgs": inputArgs, "inputArgsEncoding": inputArgsEncoding, "inputAsBytes": inputAsBytes})
        self.files = {"bofFile": ["filename.bin", bof_file]}


class TuoniCommandCd(TuoniDefaultCommand):
    def __init__(self, dir):
        super().__init__("cd", {"dir": dir})


class TuoniCommandDie(TuoniDefaultCommand):
    def __init__(self):
        super().__init__("die", {})


class TuoniCommandLs(TuoniDefaultCommand):
    def __init__(self, dir, depth = 1):
        super().__init__("ls", {"dir": dir, "depth": depth})


class TuoniCommandCmd(TuoniDefaultPluginCommand):
    def __init__(self, command, stdin = None, outputEncoding = None):
        super().__init__("cmd", {"command": command, "stdin": stdin, "outputEncoding": outputEncoding})


class TuoniCommandJobs(TuoniDefaultPluginCommand):
    def __init__(self):
        super().__init__("jobs", {})


class TuoniCommandProclist(TuoniDefaultPluginCommand):
    def __init__(self):
        super().__init__("ps", {})


class TuoniCommandRun(TuoniDefaultPluginCommand):
    def __init__(self, cmdline, output = True, stdin = None, unicode = None, outputEncoding = None):
        super().__init__("run", {"cmdline": cmdline, "output": output, "stdin": stdin, "unicode": unicode, "outputEncoding": outputEncoding})


class TuoniCommandPowershell(TuoniDefaultPluginCommand):
    def __init__(self, command = None, stdin = None, outputEncoding = None):
        super().__init__("powershell", {"command": command, "stdin": stdin, "outputEncoding": outputEncoding})


class TuoniCommandSleep(TuoniDefaultPluginCommand):
    def __init__(self, sleep_time, sleep_random):
        super().__init__("sleep", {"sleep": sleep_time, "sleepRandom": sleep_random})

#########################
## Native token commands
#########################
class TuoniCommandTokenSteal(TuoniDefaultPluginCommand):
    def __init__(self, pid):
        super().__init__("token-steal", {"pid": pid})


class TuoniCommandTokenDeleteAll(TuoniDefaultPluginCommand):
    def __init__(self):
        super().__init__("token-del-all", {})


class TuoniCommandTokenDelete(TuoniDefaultPluginCommand):
    def __init__(self, nr):
        super().__init__("token-add", {"nr": nr})


class TuoniCommandTokenList(TuoniDefaultPluginCommand):
    def __init__(self):
        super().__init__("token-list", {})


class TuoniCommandTokenMake(TuoniDefaultPluginCommand):
    def __init__(self, username, password):
        super().__init__("token-make", {"username": username, "password": password})


class TuoniCommandTokenUse(TuoniDefaultPluginCommand):
    def __init__(self, nr):
        super().__init__("token-use", {"nr": nr})


#########################
## Plugin FS commands
#########################

class TuoniCommandRm(TuoniDefaultPluginCommand):
    def __init__(self, filepath, execution_conf = None):
        super().__init__("rm", {"filepath": filepath}, execution_conf)


class TuoniCommandDownload(TuoniDefaultPluginCommand):
    def __init__(self, filepath, execution_conf = None):
        super().__init__("download", {"filepath": filepath}, execution_conf)


class TuoniCommandUpload(TuoniDefaultPluginCommand):
    def __init__(self, filepath, data, execution_conf = None):
        super().__init__("upload", {"filepath": filepath}, execution_conf)
        self.files = {"file": ["filename.bin", data]}


class TuoniCommandCp(TuoniDefaultPluginCommand):
    def __init__(self, file_from, file_to, execution_conf = None):
        super().__init__("cp", {"from": file_from, "to": file_to}, execution_conf)


class TuoniCommandMv(TuoniDefaultPluginCommand):
    def __init__(self, file_from, file_to, execution_conf = None):
        super().__init__("mv", {"from": file_from, "to": file_to}, execution_conf)


class TuoniCommandMkdir(TuoniDefaultPluginCommand):
    def __init__(self, dirpath, execution_conf = None):
        super().__init__("mkdir", {"dirpath": dirpath}, execution_conf)


#########################
## Plugin NET commands
#########################

class TuoniCommandSocks5(TuoniDefaultPluginCommand):
    def __init__(self, port, execution_conf = None):
        super().__init__("socks5", {"port": port}, execution_conf)


class TuoniCommandConnectTcp(TuoniDefaultPluginCommand):
    def __init__(self, host, port, execution_conf = None):
        super().__init__("connect-tcp", {"host": host, "port": port}, execution_conf)

#########################
## Plugin OS commands
#########################


class TuoniCommandexecuteAssembly(TuoniDefaultPluginCommand):
    def __init__(self, executable, parameters, execution_conf = None):
        super().__init__("execute-assembly", {"parameters": parameters}, execution_conf)
        self.files = {"executable": ["filename.bin", executable]}


class TuoniCommandInject(TuoniDefaultPluginCommand):
    def __init__(self, shellcode, execution_conf = None):
        super().__init__("inject", {}, execution_conf)
        self.files = {"shellcode": ["filename.bin", shellcode]}


class TuoniCommandProcinfo(TuoniDefaultPluginCommand):
    def __init__(self, execution_conf = None):
        super().__init__("procinfo", {}, execution_conf)


class TuoniCommandScreenshot(TuoniDefaultPluginCommand):
    def __init__(self, execution_conf = None):
        super().__init__("screenshot", {}, execution_conf)


class TuoniCommandSpawn(TuoniDefaultPluginCommand):
    def __init__(self, payloadId, encrypted_communication = True, execution_conf = None):
        super().__init__("spawn", {"payloadId": payloadId, "encryptedCommunication": encrypted_communication}, execution_conf)


class TuoniCommandJumpService(TuoniDefaultPluginCommand):
    def __init__(self, payloadId = None, copyMethod = None, copyPath = None, target = None, servicePath = None, serviceName = None, serviceDisplayName = None, cleanup = None, username = None, password = None, execution_conf = None):
        super().__init__("jump-service", {"payloadId":  payloadId, "copyMethod":  copyMethod, "copyPath":  copyPath, "target":  target, "servicePath":  servicePath, "serviceName":  serviceName, "serviceDisplayName":  serviceDisplayName, "cleanup":  cleanup, "username":  username, "password": password}, execution_conf)


class TuoniCommandJumpWinrm(TuoniDefaultPluginCommand):
    def __init__(self, payloadId = None, copyMethod = None, copyPath = None, target = None, executablePath = None, customPowershell = None, username = None, password = None, execution_conf = None):
        super().__init__("jump-winrm", {"payloadId":  payloadId, "copyMethod":  copyMethod, "copyPath":  copyPath, "target":  target, "executablePath":  executablePath, "customPowershell":  customPowershell, "username":  username, "password": password}, execution_conf)


class TuoniCommandJumpWmi(TuoniDefaultPluginCommand):
    def __init__(self, payloadId = None, copyMethod = None, copyPath = None, target = None, cmdline = None, username = None, password = None, execution_conf = None):
        super().__init__("jump-wmi", {"payloadId":  payloadId, "copyMethod":  copyMethod, "copyPath":  copyPath, "target":  target, "cmdline":  cmdline, "username":  username, "password": password}, execution_conf)


class TuoniCommandJumpSsh(TuoniDefaultPluginCommand):
    def __init__(self, payloadId = None, copyMethod = None, copyPath = None, target = None, cmdline = None, username = None, password = None, privateKeyPEM = None, privateKeyPassword = None, execution_conf = None):
        super().__init__("jump-wmi", {"payloadId":  payloadId, "copyMethod":  copyMethod, "copyPath":  copyPath, "target":  target, "cmdline":  cmdline, "username":  username, "password": password, "privateKeyPassword": privateKeyPassword}, execution_conf)
        self.files = {"privateKeyPEM": ["private.pem", privateKeyPEM]}

#########################
## Other commands
#########################
class TuoniCommandReverseShellCommunication(TuoniDefaultCommand):
    def __init__(self, input):
        super().__init__("cd", {"input": input})
        



import time
from tuoni.TuoniExceptions import *
from tuoni.TuoniResult import *


class TuoniCommand:
    def __init__(self, conf, c2):
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.command_id = conf["id"]
        self.configuration = conf["configuration"]
        self.execConf = conf["execConf"]
        self.created = conf["created"]
        self.sent = conf["sent"]
        self.result = conf["result"]

    def reload(self):
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get("/api/v1/commands/%d" % self.command_id)
        self._load_conf(data)

    def result_status_done(self, reload = True):
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        self.reload()
        if self.result is None:
            return False
        if self.result["status"] == "ongoing":
            return False
        return True

    def result_status_ongoing(self, reload = True):
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        self.reload()
        if self.result is None:
            return False
        if self.result["status"] == "ongoing":
            return True
        return False

    def get_result(self, reload = True):
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        self.reload()
        if self.result is None:
            return None
        return TuoniResult(self.result, self.c2)

    def wait_result(self, interval = 1, maxWait = 0):
        self.reload()
        while self.result is None:
            time.sleep(interval)
            self.reload()
            if maxWait > 0:
                maxWait -= interval
                if maxWait <= 0:
                    return None
        return TuoniResult(self.result, self.c2)




import random
from tuoni.TuoniListener import *

class TuoniPayloadPlugin:
    def __init__(self, conf, c2):
        self.name = conf["info"]["name"]
        self.vendor = conf["info"]["vendor"]
        self.description = conf["info"]["description"]
        self.plugin_id = conf["identifier"]["id"]
        self.templates = []
        for payloadTemplate in conf["payloads"]:
            self.templates.append(payloadTemplate)
        self.c2 = c2




from distutils.core import setup
setup(
  name = 'tuoni',
  packages = ['tuoni'],
  version = '0.5.0',
  license='MIT', 
  description = 'Library to interact with Tuoni C2 API',
  author = 'Shelldot',
  author_email = 'info@shelldot.com',
  url = 'https://github.com/shell-dot/tuoni-pip',
  download_url = 'https://github.com/shell-dot/tuoni-pip/archive/refs/tags/v_05_0.tar.gz',
  keywords = ['tuoni'],
  install_requires=['requests'],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12'
  ]
)